// 
//  lines.c
//  mivip
//  
//  Created by Maarten Everts.
// 

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <strings.h>
#include <string.h>
#include "lines.h"
#include "errorhandling.h"

Line *Line_new(void)
{
    return (Line*)SArray_new();
}

void Line_init(Line *self, int numPoints, int dim)
{
    int shape[2] = {numPoints,dim};
    SArray_init(self,2,shape,sizeof(float));
}

void Line_freeData(Line *self)
{
    SArray_freeData(self);
}

void Line_free(Line *self)
{
    SArray_free(self);
}


LineSeq* LineSeq_newFromTemplate(LineSeq *template)
{
    LineSeq *self = LineSeq_new();
    LineSeq_initWithCapacity(self,LineSeq_getNumLines(template));
    for(size_t i = 0; i < LineSeq_getNumLines(template); ++i) {
        Line *line_template = LineSeq_getLine(template,i);
        Line *line_new = LineSeq_addLine(self);
        Line_init(line_new,Line_getNumPoints(line_template),line_template->shape[1]);
    }
    return self;
}

void LineSeq_initCopy(LineSeq *self, LineSeq *other)
{
    LineSeq_initWithCapacity(self,LineSeq_getNumLines(other));
    for(size_t i = 0; i < LineSeq_getNumLines(other); ++i) {
        Line *line_other = LineSeq_getLine(other,i);
        Line *line_copy = LineSeq_getLine(self,i);
        Line_initCopy(line_copy,line_other);
    }
}


void LineSeq_clear(LineSeq *self)
{
    for(size_t i = 0; i < LineSeq_getNumLines(self); ++i) {
        Line *line = LineSeq_getLine(self,i);
        Line_clear(line);
    }
}

void LineSeq_free(LineSeq *self)
{
    for(size_t i = 0; i < self->size; ++i) {
        Line *l = LineSeq_getLine(self,i);
        Line_freeData(l);
    }
    DSeq_free(self);
}

void LineSeq_printInfo(LineSeq *self)
{
    int minSize=INT_MAX,maxSize=INT_MIN;
    long long totalNumPoints = 0;
    for(size_t i = 0; i < self->size; ++i) {
        Line *l = LineSeq_getLine(self,i);
        minSize = l->shape[0] < minSize ? l->shape[0] : minSize;
        maxSize = l->shape[0] > maxSize ? l->shape[0] : maxSize;
        totalNumPoints += l->shape[0];
    }
    printf("\n***** Lines info *****\n");
    printf("Number of lines: %d\n",self->size);
    printf("Mininum number of points per line: %d\n",minSize);
    printf("Maximum number of points per line: %d\n",maxSize);
    printf("Average number of points per line: %f\n",(double)totalNumPoints/(self->size));
    printf("Total number of points: %lld\n\n",totalNumPoints);
}

void Line_addAttribute(Line *self)
{
    eassertf(self->nd == 2,"SArray is not a line.");
    int newShape[2] = {self->shape[0],self->shape[1]+1};
    int newStrides[2] = {newShape[1],1};
    // allocate memory for new data
    int newNumItems = newShape[0] * newShape[1];

    float *newdata = (float*)malloc(self->itemSize*newNumItems);
    
    // copy the old data
    for(size_t i = 0; i < newShape[0]; ++i) {
        float *pointdata = Line_getPoint(self,i);
        float *target = &(newdata[newStrides[0]*i]);
        memcpy(target,pointdata,self->itemSize*self->shape[1]);
    }
    // free old memory
    free(self->data);
    // assign new memory
    self->data = newdata;
    // change shape + stride
    self->shape[0] = newShape[0]; self->shape[1] = newShape[1];
    self->strides[0] = newStrides[0]; self->strides[1] = newStrides[1];
}

void LineSeq_addAttributes(LineSeq *self)
{
    for(size_t i = 0; i < LineSeq_getNumLines(self); ++i) {
        Line_addAttribute(LineSeq_getLine(self,i));
    }
}

// int main (int argc, char const *argv[])
// {
//     printf("Testing new line data structures\n");
//     
//     LineSeq *lines = LineSeq_new();
//     LineSeq_init(lines);
//     printf("num items: %d\n",lines->size);
//     for(size_t i = 0; i < 10; ++i)
//     {
//         Line *l = LineSeq_addLine(lines);
//         Line_init(l,10,3);
//     }
//     printf("num items: %d\n",lines->size);
//     LineSeq_free(lines);
// 
// 
//     return 0;
// }
