// 
//  lines_io.c
//  mivip
//  
//  Created by Maarten Everts.
//  Copyright 2010 University of Groningen. All rights reserved.
// 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "lines_io.h"
#include "errorhandling.h"

int strendswith(const char *s1,const char *s2)
{
    size_t l1 = strlen(s1);
    size_t l2 = strlen(s2);
    return l1 > l2 && strncmp(s1 + (l1 - l2),s2,l2) == 0;
}

void LineSeq_initFromFile(LineSeq *self,const char *filename)
{
    if (strendswith(filename,".sld")) {
        LineSeq_initFromSldFile(self,filename);
    } else if (strendswith(filename,".sop")) {
        LineSeq_initFromSopFile(self,filename);
    } else {
        eprintf("File with unknown extension for line data: %s.\n",filename);
    }
}

void LineSeq_writeToFile(LineSeq *self, const char *filename)
{
    if (strendswith(filename,".sld")) {
        LineSeq_writeToSldFile(self,filename);
    } else if (strendswith(filename,".sop")) {
        LineSeq_writeToSopFile(self,filename);
    } else {
        eprintf("File with unknown extension for line data: %s.\n",filename);
    }   
}

#define MAXHEADERLINE 120

void LineSeq_initFromSldFile(LineSeq *self,const char *filename)
{
    FILE *f = fopen(filename,"rt");
    eassertf(f != NULL,"Error opening file for reading: %s\n",filename);

    char headerline[MAXHEADERLINE];
    int numLines, dim, numAttr;

    fgets(headerline, MAXHEADERLINE, f);
    if (sscanf(headerline, "%d %d %d",&numLines,&dim,&numAttr) != 3) {
        assert(sscanf(headerline, "%d\n",&numLines)==1);
        dim = 3;
        numAttr = 0;
    }
    LineSeq_initWithCapacity(self,numLines);
    for(size_t i = 0; i < numLines; ++i) {
        int numPoints;
        assert(fscanf(f,"%d\n",&numPoints)==1);
        Line *l = LineSeq_addLine(self);
        Line_init(l,numPoints,dim+numAttr);
        for(size_t p = 0; p < numPoints; ++p) {
            float *point = Line_getPoint(l,p);
            for(size_t j = 0; j < dim+numAttr; ++j) {
                assert(fscanf(f,"%f",&(point[j]))==1);
            }
        }
    }
    fclose(f);
}

#define LineSeq_DEFAULTDIM 3
void LineSeq_writeToSldFile(LineSeq *self, const char *filename)
{
    FILE *f = fopen(filename,"wt");
    eassertf(f != NULL,"Error opening file for writing: %s\n",filename);
    int numAttr = LineSeq_getLine(self,0)->shape[1] - LineSeq_DEFAULTDIM;
    fprintf(f,"%d %d %d\n",LineSeq_getNumLines(self),LineSeq_DEFAULTDIM,numAttr);
    for(size_t i = 0; i < LineSeq_getNumLines(self); ++i) {
        Line *line = LineSeq_getLine(self,i);
        fprintf(f,"%d\n",Line_getNumPoints(line));
        for(size_t p = 0; p < Line_getNumPoints(line); ++p) {
            float *point = Line_getPoint(line,p);
            fprintf(f,"%.9g",point[0]);
            for(size_t j = 1; j < line->shape[1]; ++j) {
                fprintf(f," %.9g",point[j]);
            }
            fprintf(f,"\n");
        }
    }
    fclose(f);
}

typedef struct {
	unsigned char version; // 0 :)
	unsigned char endianness; // assumption for this code: little endian
	unsigned int endiantest; // to test/check endianness
	unsigned int dim; // assumption: 3
	unsigned char dtype; // unused, we're assuming float for now
	unsigned int numSequences;
	unsigned int dataStartByte;
	unsigned int numAttributes; // assumption for now: 1
	unsigned int attributesSize; // per point, in bytes (assumption: sizeof(float)).
	/* After this: optional attributes description */
} SopFileHeader;

#define SopFile_ENDIANTESTVALUE 14600958

void readSopHeader(FILE *f, SopFileHeader *header)
{
	int itemsRead = fread(header, sizeof(SopFileHeader), 1, f);
	assert(itemsRead==1);
	assert(header->version==0);
	assert(header->endianness==0);
	assert(header->endiantest==SopFile_ENDIANTESTVALUE);
	if (header->numAttributes > 0 ) {
		assert(header->attributesSize==header->numAttributes*sizeof(float)); 
	}
}

void LineSeq_initFromSopFile(LineSeq *self, const char *filename)
{
    FILE *f = fopen(filename, "rb");
    eassertf(f != NULL,"Error opening file for reading: %s\n",filename);
    SopFileHeader header;
    readSopHeader(f,&header);
    fseek(f, header.dataStartByte, SEEK_SET);
    LineSeq_initWithCapacity(self,header.numSequences);
    int totalDim = header.dim + header.numAttributes;
    for(size_t i = 0; i < header.numSequences; ++i) {
        Line *line = LineSeq_addLine(self);
        int numPoints;
        fread(&numPoints, sizeof(int), 1, f);
        Line_init(line, numPoints, totalDim);
        if (header.numAttributes == 0) {
            int numRead = fread(line->data, sizeof(float), totalDim*numPoints, f);
            assert(numRead == totalDim*numPoints);
        } else {
            // Why did I put the attributes separate in the file?!
            // Ah well, have to deal with it.
            {
                float *tmpData = (float*)malloc(sizeof(float)*header.dim*numPoints);
                int numRead = fread(tmpData,sizeof(float),header.dim*numPoints,f);
                assert(numRead=header.dim*numPoints);
                // copy position data to line
                for(size_t p = 0; p < numPoints; ++p) {
                    float *point = Line_getPoint(line,p);
                    memcpy(point,tmpData+header.dim*p,sizeof(float)*header.dim);
                }
                free(tmpData);
            }
            {
                float *tmpData = (float*)malloc(sizeof(float)*header.numAttributes*numPoints);
                int numRead = fread(tmpData,sizeof(float),header.numAttributes*numPoints,f);
                assert(numRead=header.numAttributes*numPoints);
                // copy attribute data to line
                for(size_t p = 0; p < numPoints; ++p) {
                    float *point = Line_getPoint(line,p) + header.dim;
                    memcpy(point,tmpData+header.numAttributes*p,sizeof(float)*header.numAttributes);
                }                
                free(tmpData);
            }
        }
    }
    fclose(f);
}

void LineSeq_writeToSopFile(LineSeq *self, const char *filename)
{
	FILE *f = fopen(filename, "wb");
    eassertf(f != NULL,"Error opening file for writing: %s\n",filename);
	// fill a header
    int numAttr = LineSeq_getLine(self,0)->shape[1] - LineSeq_DEFAULTDIM;
    SopFileHeader header = {0x00,  //version
                            0x00,  //endianness
                            SopFile_ENDIANTESTVALUE, //endiantest
                            LineSeq_DEFAULTDIM, //dim
                            0x00, // dtype
                            LineSeq_getNumLines(self),//numSequences
                            sizeof(SopFileHeader) + 3,//dataStartByte
                            numAttr, //numAttr
                            sizeof(float) //attributesSize
    };
                  
    // write the header
    size_t itemsWritten = fwrite(&header,sizeof(SopFileHeader),1,f);
    eassertf(itemsWritten == 1,"Error writing header to .sop file\n");
    fseek(f,header.dataStartByte,SEEK_SET);
    
	// write the lines
	for(size_t i = 0; i < LineSeq_getNumLines(self); ++i) {
        Line *line = LineSeq_getLine(self,i);
        int numPoints = Line_getNumPoints(line);
        fwrite(&(numPoints),sizeof(int),1,f);
        if (numAttr == 0) {
            fwrite(line->data,sizeof(float),numPoints*LineSeq_DEFAULTDIM,f);
        } else {
            // Again, the attribute data comes after the position data...
            // Not very convenient for this dataset
            float *tmpPosData = (float*)malloc(sizeof(float)*numPoints*LineSeq_DEFAULTDIM);
            float *tmpAttrData = (float*)malloc(sizeof(float)*numPoints*numAttr);
            for(size_t p = 0; p < numPoints; ++p) {
                float *point = Line_getPoint(line,p);
                memcpy(tmpPosData + LineSeq_DEFAULTDIM*p,point,sizeof(float)*LineSeq_DEFAULTDIM);
                memcpy(tmpAttrData + numAttr*p,point+LineSeq_DEFAULTDIM,sizeof(float)*numAttr);
            }
            fwrite(tmpPosData,sizeof(float),numPoints*LineSeq_DEFAULTDIM,f);
            fwrite(tmpAttrData,sizeof(float),numPoints*numAttr,f);
            free(tmpAttrData);
            free(tmpPosData);
        }
	}
    fclose(f);
}

