#!/usr/bin/env python
"""
linesfile.py

A simple module for reading and writing .lines files. A .lines file consists
of two parts: a json-formatted header (1st line) and the line data.
The data can be text or binary.
"""

import numpy as np
import json
import struct

def __write_text_linedata__(f,linedata):
    for line in linedata:
        f.write(str(line.shape[0])+"\n")
        for point in line:
            f.write(' '.join([str(c) for c in point])+"\n")

def __write_binary_linedata__(f,linedata):
    for line in linedata:
        f.write(struct.pack('<l', line.shape[0]))
        line.tofile(f)

def write(linedata,filename,dim=3,format="text",attributes=[],extent=[]):
    """Write linedata (list of numpy arrays) to a .lines file."""
    num_lines = len(linedata)
    num_attr = linedata[0].shape[1] - dim
    attributes_n = list(attributes)
    if len(attributes_n) < num_attr:
        attributes_n += ["" for _ in range(num_attr - len(attributes_n))]
    elif len(attributes_n) > num_attr:
        print "Warning: number of attribute descriptions too large."
        attributes_n = attributes_n[:num_attr]
    header = {"dim": dim,
              "num_lines": num_lines,
              "attributes": attributes_n,
              "format": format
              }
    if len(extent) > 0:
        header['extent'] = extent
    header_string = json.dumps(header)
    with open(filename,'w') as f:
        f.write(header_string+"\n")
        if header['format'] == 'text':
            __write_text_linedata__(f,linedata)
        elif header['format'] == 'binary':
            __write_binary_linedata__(f,linedata)
        else:
            print "Unknown format!"

def __read_binary_linedata__(f,header):
    num_scalars = header['dim'] + len(header['attributes'])
    lines = []
    for _ in range(header['num_lines']):
        num_points = struct.unpack("<l",f.read(4))[0]
        lines.append(np.fromfile(f,np.float32,num_points*num_scalars).reshape((num_points,num_scalars)))
    return lines

def __read_text_linedata__(f,header):
    lines = []
    text_lines = f.readlines()
    c = 0
    for _ in range(header['num_lines']):
        num_points = int(text_lines[c].strip())
        lines.append(np.array(
                [[float(x) for x in p.strip().split(' ')] 
                 for p in text_lines[c+1:c+1+num_points]],
                np.float32))
        c += num_points + 1
    return lines

def read(filename):
    """Reads line data from .lines file named filename and returns
    a tuple with the header and the line data (list of numpy arrays).
    """
    with open(filename,'rt') as f:
        header_line = f.readline()
        header = json.loads(header_line)
        if header['format'] == 'text':
            linedata = __read_text_linedata__(f,header)
        elif header['format'] == 'binary':
            linedata = __read_binary_linedata__(f,header)
        return header,linedata
