#!/usr/bin/env python
# encoding: utf-8
"""
sopfile.py

Simple method for reading a .sop file (sequence of points)
into a list of numpy arrays.

Created by Maarten Everts on 2009-07-07.
Copyright (c) 2009 University of Groningen. All rights reserved.
"""

import numpy as np
import struct

__ENDIANTESTVALUE__ = 14600958

__sopfileheader__ = np.dtype([
    ('version',np.byte,1), # 0 :)
    ('endianness',np.byte,1), # assumption for this code: little endian
    ('endiantest',np.uint32,1), # to test/check endianness
    ('dim',np.uint32,1), # assumption: 3
    ('dtype',np.byte,1), # unused, we're assuming float for now
    ('numSequences',np.uint32,1), 
    ('dataStartByte',np.uint32,1),
    ('numAttributes',np.uint32,1), # assumption for now: 0 or 1
    ('attributesSize',np.uint32,1) # per point, in bytes 
                                   # (assumption: sizeof(float)).
    # After this: optional attributes description
],True)

def read(filename):
    """Read a .sop file, returns a list of numpy arrays."""
    infile = open(filename,'rb')
    header = np.fromfile(infile, dtype=__sopfileheader__, count=1)[0]
    assert header['version'] == 0, "Only version 0 is supported."
    assert header['endianness'] == 0, "Only little endian is supported."
    assert header['endiantest'] == __ENDIANTESTVALUE__, "Endian test incorrect."
    scalars_per_point = 3+header['numAttributes']
    lines = []
    infile.seek(header['dataStartByte'])
    for i in range(header['numSequences']):
        num_points = struct.unpack("<l", infile.read(4))[0]
        line = np.fromfile(infile, np.float32, num_points*scalars_per_point)
        if scalars_per_point > 3:
            # In the sopfile, the attribute data is stored *after* the position data
            # so we need to reorder the data.
            lines.append(np.hstack([line[:num_points*3].reshape(num_points,3),
                                    line[num_points*3:].reshape(num_points,1)]))
        else:
            # No attribute (extra) data, so just take the data as is and simply
            # reshape it.
            lines.append(line.reshape(num_points, scalars_per_point))
    return lines

def write(data, filename):
    """Save data to .sop file with name filename."""
    # First check whether the data is in the expected form
    assert len(data) > 0
    assert data[0].shape[1] >= 3
    assert all(t.shape[1]==data[0].shape[1] for t in data)
    num_attr = data[0].shape[1] - 3
    outfile = open(filename,'wb')
    header = np.array([(0, 0, __ENDIANTESTVALUE__, 3, 1,
                      len(data), __sopfileheader__.itemsize,
                      num_attr, 4)], dtype=__sopfileheader__)
    header.tofile(outfile)
    for line in data:
        outfile.write(struct.pack("<l", line.shape[0]))
        if num_attr > 0:
            line[:,:3].tofile(outfile)
            line[:,3:].tofile(outfile)
        else:
            line.tofile(outfile)
    outfile.close()

