#!/usr/bin/env python
# encoding: utf-8
"""
trkdata.py

A simple class for reading TrackVis .trk files. Tracts are stored
in a list of of numpy arrays.

Created by Maarten Everts on 2009-07-07.
Copyright (c) 2009 University of Groningen. All rights reserved.
"""

import numpy as np
import struct

__trkfileheader__ = np.dtype([
    ('id_string', np.str_, 6),
    ('dim', np.int16, (3,)),
    ('voxel_size',np.float32,(3,)),
    ('origin',np.float32,(3,)),
    ('n_scalars',np.int16,1),
    ('scalar_name',np.str_,200),
    ('n_properties',np.int16,1),
    ('property_name',np.str_,200),
    ('reserved',np.str_,508),
    ('voxel_order',np.byte,4),
    ('pad2',np.byte,4),
    ('image_orientation_patient',np.float32,(6,)),
    ('pad1',np.byte,2),
    ('invert_x',np.byte,1),
    ('invert_y',np.byte,1),
    ('invert_z',np.byte,1),
    ('swap_xy',np.byte,1),
    ('swap_yz',np.byte,1),
    ('swap_zx',np.byte,1),
    ('n_count',np.int32,1),
    ('version',np.int32,1),
    ('hdr_size',np.int32,1)
])


class Trkfile():
    """Class for reading TrackVis .trk files."""
    def __init__(self, filename, rescale=False):
        infile = open(filename,'rb')
        self.header = np.fromfile(infile, dtype=__trkfileheader__, count=1)[0]
        scalarsperpoint = 3 + self.header['n_scalars']
        self.tracts = []
        # Read the number of points as binary in a string, unpack later
        num_points_s = infile.read(4)
        # Read tracts until EOF
        while len(num_points_s) > 0:
            num_points = struct.unpack("<l", num_points_s)[0]
            tract = np.fromfile(infile, np.float32, num_points*scalarsperpoint)
            self.tracts.append(tract.reshape(num_points, scalarsperpoint))
            if self.header['n_properties'] > 0:
                # For now, just ignore these tract properties
                properties = infile.read(self.header['n_properties']*4)
            num_points_s = infile.read(4)
        infile.close()
        if rescale:
            self.rescale()
    def rescale(self):
        """Rescale tract positions according to voxel size."""
        voxel_size = self.header['voxel_size']
        for tract in self.tracts:
            tract[:, 0] *= voxel_size[0]
            tract[:, 1] *= voxel_size[1]
            tract[:, 2] *= voxel_size[2]


